"""Img2dataset"""

from multiprocessing import Pool
from typing import List, Optional
from tqdm import tqdm
import fire
import logging
from .logger import LoggerProcess
from .resizer import Resizer
from .writer import WebDatasetSampleWriter, FilesSampleWriter, ParquetSampleWriter
from .reader import Reader
from .downloader import Downloader
import fsspec
import sys
import signal

logging.getLogger("exifread").setLevel(level=logging.CRITICAL)


def download(
    url_list: str,
    image_size: int = 256,
    output_folder: str = "images",
    processes_count: int = 1,
    resize_mode: str = "border",
    resize_only_if_bigger: bool = False,
    upscale_interpolation: str = "lanczos",
    downscale_interpolation: str = "area",
    encode_quality: int = 95,
    skip_reencode: bool = False,
    output_format: str = "files",
    input_format: str = "txt",
    url_col: str = "url",
    caption_col: Optional[str] = None,
    thread_count: int = 256,
    number_sample_per_shard: int = 10000,
    extract_exif: bool = True,
    save_additional_columns: Optional[List[str]] = None,
    timeout: int = 10,
    enable_wandb: bool = False,
    wandb_project: str = "img2dataset",
    oom_shard_count: int = 5,
    compute_md5: bool = True,
):
    """Download is the main entry point of img2dataset, it uses multiple processes and download multiple files"""
    config_parameters = dict(locals())
    logger_process = LoggerProcess(output_folder, enable_wandb, wandb_project, config_parameters, processes_count)
    logger_process.start()

    def signal_handler(signal_arg, frame):  # pylint: disable=unused-argument
        logger_process.terminate()
        sys.exit(0)

    signal.signal(signal.SIGINT, signal_handler)

    save_caption = caption_col is not None

    fs, output_path = fsspec.core.url_to_fs(output_folder)

    if not fs.exists(output_path):
        fs.mkdir(output_path)
        start_shard_id = 0
    else:
        existing_top_level_files = fs.glob(output_path + "/*")
        if len(existing_top_level_files) == 0:
            start_shard_id = 0
        else:
            start_shard_id = max([int(x.split("/")[-1].split(".")[0]) for x in existing_top_level_files]) + 1

    reader = Reader(
        url_list, input_format, url_col, caption_col, save_additional_columns, number_sample_per_shard, start_shard_id
    )

    if output_format == "webdataset":
        sample_writer_class = WebDatasetSampleWriter
    elif output_format == "parquet":
        sample_writer_class = ParquetSampleWriter  # type: ignore
    elif output_format == "files":
        sample_writer_class = FilesSampleWriter  # type: ignore

    resizer = Resizer(
        image_size=image_size,
        resize_mode=resize_mode,
        resize_only_if_bigger=resize_only_if_bigger,
        upscale_interpolation=upscale_interpolation,
        downscale_interpolation=downscale_interpolation,
        encode_quality=encode_quality,
        skip_reencode=skip_reencode,
    )

    downloader = Downloader(
        sample_writer_class=sample_writer_class,
        resizer=resizer,
        thread_count=thread_count,
        save_caption=save_caption,
        extract_exif=extract_exif,
        output_folder=output_folder,
        column_list=reader.column_list,
        timeout=timeout,
        number_sample_per_shard=number_sample_per_shard,
        oom_shard_count=oom_shard_count,
        compute_md5=compute_md5,
    )

    print("Starting the downloading of this file")
    with Pool(processes_count, maxtasksperchild=5) as process_pool:
        for _ in tqdm(process_pool.imap_unordered(downloader, reader),):
            pass

        process_pool.terminate()
        process_pool.join()
        del process_pool
    logger_process.join()


def main():
    fire.Fire(download)


if __name__ == "__main__":
    main()



"""resizer module handle image resizing"""

import albumentations as A
import cv2
import numpy as np
from enum import Enum
import imghdr

_INTER_STR_TO_CV2 = dict(
    nearest=cv2.INTER_NEAREST,
    linear=cv2.INTER_LINEAR,
    bilinear=cv2.INTER_LINEAR,
    cubic=cv2.INTER_CUBIC,
    bicubic=cv2.INTER_CUBIC,
    area=cv2.INTER_AREA,
    lanczos=cv2.INTER_LANCZOS4,
    lanczos4=cv2.INTER_LANCZOS4,
)


class ResizeMode(Enum):
    no = 0
    keep_ratio = 1
    center_crop = 2
    border = 3


def inter_str_to_cv2(inter_str):
    inter_str = inter_str.lower()
    if inter_str not in _INTER_STR_TO_CV2:
        raise Exception(f"Invalid option for interpolation: {inter_str}")
    return _INTER_STR_TO_CV2[inter_str]


class Resizer:
    """
    Resize images
    Expose a __call__ method to be used as a callable object

    Should be used to resize one image at a time

    Options:
        resize_mode: "no", "keep_ratio", "center_crop", "border"
        resize_only_if_bigger: if True, resize only if image is bigger than image_size
        image_size: size of the output image to resize
    """

    def __init__(
        self,
        image_size,
        resize_mode,
        resize_only_if_bigger,
        upscale_interpolation="lanczos",
        downscale_interpolation="area",
        encode_quality=95,
        skip_reencode=False,
    ):
        self.image_size = image_size
        if isinstance(resize_mode, str):
            if resize_mode not in ResizeMode.__members__:  # pylint: disable=unsupported-membership-test
                raise Exception(f"Invalid option for resize_mode: {resize_mode}")
            resize_mode = ResizeMode[resize_mode]
        self.resize_mode = resize_mode
        self.resize_only_if_bigger = resize_only_if_bigger
        self.upscale_interpolation = inter_str_to_cv2(upscale_interpolation)
        self.downscale_interpolation = inter_str_to_cv2(downscale_interpolation)
        self.encode_params = [int(cv2.IMWRITE_JPEG_QUALITY), encode_quality]
        self.skip_reencode = skip_reencode

    def __call__(self, img_stream):
        """
        input: an image stream
        output: img_str, width, height, original_width, original_height, err
        """
        try:
            encode_needed = imghdr.what(img_stream) != "jpeg" if self.skip_reencode else True
            img_buf = np.frombuffer(img_stream.read(), np.uint8)
            img = cv2.imdecode(img_buf, cv2.IMREAD_UNCHANGED)
            if img is None:
                raise Exception("Image decoding error")
            if len(img.shape) == 3 and img.shape[-1] == 4:
                # alpha matting with white background
                alpha = img[:, :, 3, np.newaxis]
                img = alpha / 255 * img[..., :3] + 255 - alpha
                img = np.rint(img.clip(min=0, max=255)).astype(np.uint8)
                encode_needed = True
            original_height, original_width = img.shape[:2]

            # resizing in following conditions
            if self.resize_mode in (ResizeMode.keep_ratio, ResizeMode.center_crop):
                downscale = min(original_width, original_height) > self.image_size
                if not self.resize_only_if_bigger or downscale:
                    interpolation = self.downscale_interpolation if downscale else self.upscale_interpolation
                    img = A.smallest_max_size(img, self.image_size, interpolation=interpolation)
                    if self.resize_mode == ResizeMode.center_crop:
                        img = A.center_crop(img, self.image_size, self.image_size)
                    encode_needed = True
            elif self.resize_mode == ResizeMode.border:
                downscale = max(original_width, original_height) > self.image_size
                if not self.resize_only_if_bigger or downscale:
                    interpolation = self.downscale_interpolation if downscale else self.upscale_interpolation
                    img = A.longest_max_size(img, self.image_size, interpolation=interpolation)
                    img = A.pad(
                        img, self.image_size, self.image_size, border_mode=cv2.BORDER_CONSTANT, value=[255, 255, 255]
                    )
                    encode_needed = True
            height, width = img.shape[:2]
            if encode_needed:
                img_str = cv2.imencode(".jpg", img, params=self.encode_params)[1].tobytes()
            else:
                img_str = img_buf.tobytes()
            return img_str, width, height, original_width, original_height, None

        except Exception as err:  # pylint: disable=broad-except
            return None, None, None, None, None, str(err)



"""Reader is module to read the url list and return shards"""

import pandas as pd
import tempfile
import math
import fsspec


class Reader:
    """
    The reader class reads an url list and returns shards
    It provides an iter method
    It provides attributes:
    - column_list: the list of columns to read
    - input_format: the format of the input file
    - url_col: the column name of the url
    - caption_col: the column name of the caption
    - save_additional_columns: the list of additional columns to save
    - number_sample_per_shard: the number of samples per shard
    - start_shard_id: the id of the first shard
    """

    def __init__(
        self,
        url_list,
        input_format,
        url_col,
        caption_col,
        save_additional_columns,
        number_sample_per_shard,
        start_shard_id,
    ) -> None:
        self.input_format = input_format
        self.url_col = url_col
        self.caption_col = caption_col
        self.save_additional_columns = save_additional_columns
        self.number_sample_per_shard = number_sample_per_shard
        self.start_shard_id = start_shard_id

        fs, url_path = fsspec.core.url_to_fs(url_list)
        self.fs = fs

        if fs.isdir(url_path):
            self.input_files = sorted(fs.glob(url_path + "/*." + input_format))
        else:
            self.input_files = [url_path]

        if self.input_format == "txt":
            self.column_list = ["url"]
        elif self.input_format in ["json", "csv", "tsv", "tsv.gz", "parquet"]:
            self.column_list = self.save_additional_columns if self.save_additional_columns is not None else []
            if self.caption_col is not None:
                self.column_list = self.column_list + ["caption", "url"]
            else:
                self.column_list = self.column_list + ["url"]

    def _save_to_arrow(self, input_file, tmp_dir):
        """Read the input file and save to arrow files in a temporary directory"""
        if self.input_format in ["txt", "json", "csv", "tsv"]:
            with self.fs.open(input_file, encoding="utf-8", mode="r") as file:
                if self.input_format == "txt":
                    df = pd.DataFrame([(url.rstrip(),) for url in file.readlines()], columns=self.column_list)
                elif self.input_format == "json":
                    df = pd.read_json(file)
                elif self.input_format == "csv":
                    df = pd.read_csv(file)
                elif self.input_format == "tsv":
                    df = pd.read_table(file)
        elif self.input_format in ["tsv", "tsv.gz", "parquet"]:
            with self.fs.open(input_file, mode="rb") as file:
                if self.input_format == "tsv.gz":
                    df = pd.read_table(file, compression="gzip")
                elif self.input_format == "parquet":
                    columns_to_read = [self.url_col]
                    if self.caption_col is not None:
                        columns_to_read += [self.caption_col]
                    if self.save_additional_columns is not None:
                        columns_to_read += self.save_additional_columns
                    df = pd.read_parquet(file, columns=columns_to_read)
        else:
            assert False, f"Unexpected input format ({self.input_format})."

        df = df.rename(columns={self.caption_col: "caption", self.url_col: "url"})
        df = df.where(pd.notnull(df), None)

        number_samples = len(df)

        number_shards = math.ceil(len(df) / self.number_sample_per_shard)

        shards = []
        for shard_id in range(number_shards):
            begin_shard = shard_id * self.number_sample_per_shard
            end_shard = min(number_samples, (1 + shard_id) * self.number_sample_per_shard)
            df_shard = df[begin_shard:end_shard][self.column_list]
            df_shard = df_shard.reset_index(drop=True)
            file_name = tmp_dir + f"/{shard_id + self.start_shard_id}.feather"
            df_shard.to_feather(file_name)
            shards.append((shard_id, file_name))
        del df

        return shards

    def __iter__(self):
        """
        Iterate over shards, yield shards of size number_sample_per_shard or less for the last one
        Each shard is a tuple (shard_id, shard)
        shard is a tuple (sample id, sample)
        sample is a tuple of the columns
        """
        for i, input_file in enumerate(self.input_files):
            print(
                "Downloading file number " + str(i + 1) + " of " + str(len(self.input_files)) + " called " + input_file
            )
            print("Loading the input file")

            with tempfile.TemporaryDirectory("img2dataset") as tmp_dir:
                shards = self._save_to_arrow(input_file, tmp_dir)
                num_shard = 0
                for num_shard, arrow_file in shards:
                    df = pd.read_feather(arrow_file)
                    images_to_dl = df[self.column_list].to_records(index=False).tolist()
                    del df

                    yield (
                        num_shard + self.start_shard_id,
                        list(enumerate(images_to_dl)),
                    )

                    num_shard += 1
                self.start_shard_id += num_shard
                del images_to_dl



from setuptools import setup, find_packages
from pathlib import Path

if __name__ == "__main__":
    with Path(Path(__file__).parent, "README.md").open(encoding="utf-8") as file:
        long_description = file.read()

    setup(
        name="img2dataset",
        packages=find_packages(),
        include_package_data=True,
        version="1.18.0",
        license="MIT",
        description="Easily turn a set of image urls to an image dataset",
        long_description=long_description,
        long_description_content_type="text/markdown",
        entry_points={"console_scripts": ["img2dataset = img2dataset:main"]},
        author="Romain Beaumont",
        author_email="romain.rom1@gmail.com",
        url="https://github.com/rom1504/img2dataset",
        data_files=[(".", ["README.md"])],
        keywords=["machine learning", "computer vision", "download", "image", "dataset"],
        install_requires=[
            "tqdm",
            "opencv-python",
            "fire",
            "webdataset",
            "pandas",
            "pyarrow",
            "exifread",
            "albumentations",
            "pyyaml",
            "dataclasses",
            "wandb",
            "fsspec",
        ],
        classifiers=[
            "Development Status :: 4 - Beta",
            "Intended Audience :: Developers",
            "Topic :: Scientific/Engineering :: Artificial Intelligence",
            "License :: OSI Approved :: MIT License",
            "Programming Language :: Python :: 3.6",
        ],
    )
